/*
rk4.cc

This C++ file defines member functions for the class RK4_MODULE
It was generated with rcsdesign
with template version 1.1

MODIFICATIONS:
12-Sep-99	Created with RCS-Design tool.

*/

// Include Files
#include "rcs.hh" 	// Common RCS definitions
#include "nml_mod.hh" 	// NML_MODULE definitions
#include "rk4.hh" 	// RK4_MODULE definitions
// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.


#include "rk4n.hh" 	// NML Commands and Status definitions for rk4
// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


// Constructor
RK4_MODULE::RK4_MODULE(int _is_base_class = 0)
  : RK4_SIM(PLANT_ORDER, CTRL_ORDER)
{

	// Initialize the NML channels if this module is not being used as the base class for another module.
	if(!_is_base_class)
	{
		INITIALIZE_NML();
	}
	// Add additional code to initialize the module here.
  if (rk4_sim_err != RK4_SIM_NO_ERROR)
    {
      logError("Error in creating RK4_MODULE object!!!");
      status = RCS_ERROR; // let my supervisor know I have a problem.
    }

}

// Overloaded Virtual Functions

/*
INITIALIZE_NML

The INITIALIZE_NML function is ussually called only once from within the
constructor. It should not be called if a derived class will also call it.

*/
void RK4_MODULE::INITIALIZE_NML()
{
	// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	setErrorLogChannel(new NML(nmlErrorFormat, "errlog", "rk4", "rk4_module.nml"));
	setCmdChannel(new RCS_CMD_CHANNEL(rk4Format, "rk4_cmd", "rk4", "rk4_module.nml"));
	rk4_status = new RK4_STATUS();
	setStatChannel(new RCS_STAT_CHANNEL(rk4Format, "rk4_sts", "rk4", "rk4_module.nml"), rk4_status);
	// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


}

/*
PRE_PROCESS

The PRE_PROCESS function is called every cycle after the command and
subordinates status have been read but before DECISION_PROCESS is called.
It is intended to be used for tasks such as sensory processing that should
be performed every cycle regardless of the current command or state.
*/
void RK4_MODULE::PRE_PROCESS()
{
	// Pre-Processing Code
}

/*
DECISION_PROCESS

The DECISION_PROCESS function is called every cycle as long as there is a non-zero command.
It is expected to call a command function based on commandInData->type.
*/
void RK4_MODULE::DECISION_PROCESS()
{
	switch(commandInData->type)
	{
		// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	case RK4_INIT_TYPE:
		INIT((RK4_INIT *)commandInData);
		break;

	case RK4_HALT_TYPE:
		HALT((RK4_HALT *)commandInData);
		break;

	case RK4_RUN_TYPE:
		RUN((RK4_RUN *)commandInData);
		break;
		// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


	default:
		logError("The command %d is not recognized.",commandInData->type);
		break;
	}
}

/*
POST_PROCESS

The POST_PROCESS function is called every cycle after DECISION_PROCESS is called
but before the status and the subordinates commands  have been written.
It is intended to be used for tasks such as output filters that should
be performed every cycle regardless of the current command or state.
*/
void RK4_MODULE::POST_PROCESS()
{
	// Post-Processing Code
  rk4_status->current_time = current_time;
  memcpy(rk4_status->plant_state, plant_state, plant_order*sizeof(double));
  memcpy(rk4_status->ctrl_input, ctrl_input, ctrl_order*sizeof(double));
}

// Command Functions

/*
INIT

Parameter(s):
RK4_INIT *cmd_in -- NML Message sent from superior.

Most Modules will have an INIT command.
The INIT function is expected to initialize any variables that may be
in an uninitialized or unknown state, send INIT commands to its subordinates,
wait for the subordinates to be DONE and then inform its superior that it is done.
The state tables should use the STATE_MATCH macro so the diagnostics tool can 
highlight the current line in the state table.
*/
void RK4_MODULE::INIT(RK4_INIT *cmd_in)
{
	// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


	if(STATE_MATCH(NEW_COMMAND))
	{
		// Send an INIT command to all subordinates.
		// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

		// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


		stateNext(S1);
		// Reinitialize variables here.

	}
	// Wait for all subordinates to report done.
	// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	else if(STATE_MATCH(S1,
		1))
	// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.

	{
	  cycle_time = cmd_in->cycle_time;
	  status = RCS_DONE;
	  stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}

/*
HALT

Parameter(s):
RK4_HALT *cmd_in -- NML Message sent from superior.

Most Modules will have an HALT command.
The HALT function is expected to stop any motion or command execution,
send HALT commands to its subordinates,
wait for the subordinates to be DONE and then inform its superior that it is done.
The state tables should use the STATE_MATCH macro so the diagnostics tool can 
highlight the current line in the state table.
*/
void RK4_MODULE::HALT(RK4_HALT *cmd_in)
{
	// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


	if(STATE_MATCH(NEW_COMMAND))
	{
		//Send a HALT command to all subordinates.
		// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

		// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.

		stateNext(S1);
	}
	// Wait for all subordinates to report done.
	// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	else if(STATE_MATCH(S1,
		1))
	// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.

	{
		status = RCS_DONE;
		stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}

/*
RUN

Parameter(s):
RK4_RUN *cmd_in -- NML Message sent from superior.

*/
void RK4_MODULE::RUN(RK4_RUN *cmd_in)
{
	// Put state table for RK4_SIM here.
  if(STATE_MATCH(NEW_COMMAND))
    {
      num_sim_per_cycle = cmd_in->num_sim_per_cycle;
      init_param(cycle_time/num_sim_per_cycle, 
		 cmd_in->initial_time,
		 cmd_in->initial_state);
      status = RCS_EXEC;
      stateNext(S1);
    }
  else if(STATE_MATCH(S1))
    {
      if(current_time >= cmd_in->final_time)
	{
	  status = RCS_DONE;
	  stateNext(S2);
	}
      else
	{
	  // Calculate the control input here before calling the 
	  // simulation function
	  FIND_CTRL_INPUT();

	  // Simulate the system
	  for (int ii=0; ii < num_sim_per_cycle; ii++)
	    {
	      solve();
	    }
	}
    }
  else if(STATE_MATCH(S2))
    {
      // Idle State
    }
  
}

void RK4_MODULE::DERIV (const double t, const double *x,
                        const double *u, double *xdot)
{
  // This is where the derivative should be defined in the
  // form: xdot = f(t,x,u), where t is the time, x the current state,
  // u is the current control input, and xdot is the derivative to be
  // computed here
  
  // For example for automobile model with
  // Simple point mass dynamics, states
  // x[0] = longitudinal velocity
  // x[1] = lateral position
  // x[2] = lateral velocity
  // and parameters: mass, drag_x, drag_y
  // The equations are as shown below

  // xdot[0] = (1.0/mass)*(-drag_x*(x[0] + u[2]) + u[0]*cos(u[1]));
  // xdot[1] = x[2];
  // xdot[2] = (1.0/mass)*(-drag_y*x[2] + u[0]*sin(u[1]));
  
}

void RK4_MODULE::FIND_CTRL_INPUT(void)
{
  // Calculate the values of the control input ctrl_input[ii]

  //  for (int ii=0; ii < ctrl_order; ii++)
  //  {
  //    ctrl_input[ii] = 0.0;
  //  }

}

